#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef _MPI
	#include <mpi.h>
#endif
#include "lis.h"
#include "lis_matrix.h"

#define NNZ_BLOCK	1000

int lis_matrix_create(int local_n, int global_n, LIS_Comm comm, LIS_MATRIX *Amat)
{
	int n,gn;
	int nprocs,my_rank;
	int is,ie;
	int i;
	int *ranges;

	*Amat = NULL;

	if( global_n>0 && local_n>global_n )
	{
		SETERR("local n(=%d) is larger than global n(=%d)\n",local_n,global_n);
		return LIS_FAILS;
	}
	if( local_n<0 || global_n<0 )
	{
		SETERR("local n(=%d) or global n(=%d) are less than 0\n",local_n,global_n);
		return LIS_FAILS;
	}
	if( local_n==0 && global_n==0 )
	{
		SETERR("local n(=%d) and global n(=%d) are 0\n",local_n,global_n);
		return LIS_FAILS;
	}

	*Amat = (LIS_MATRIX)lis_malloc( sizeof(struct LIS_MATRIX_STRUCT) );
	if( NULL==*Amat )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	#ifdef _MPI
		MPI_Comm_size(comm,&nprocs);
		MPI_Comm_rank(comm,&my_rank);
		ranges = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
		if( ranges==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
	#else
		nprocs  = 1;
		my_rank = 0;
		ranges  = NULL;
	#endif

	if( local_n==0 )
	{
		#ifdef _MPI
			LIS_GET_ISIE(my_rank,nprocs,global_n,is,ie);
			local_n = ie-is;
			MPI_Allgather(&ie,1,MPI_INT,&ranges[1],1,MPI_INT,comm);
			ranges[0] = 0;
		#else
			local_n = global_n;
			is      = 0;
			ie      = global_n;
		#endif
	}
	else
	{
		#ifdef _MPI
			MPI_Allgather(&local_n,1,MPI_INT,&ranges[1],1,MPI_INT,comm);
			ranges[0] = 0;
			for(i=0;i<nprocs;i++)
			{
				ranges[i+1] += ranges[i];
			}
			global_n = ranges[nprocs];
			is       = ranges[my_rank];
			ie       = ranges[my_rank+1];
		#else
			global_n = local_n;
			is       = 0;
			ie       = local_n;
		#endif
	}

	(*Amat)->value       = NULL;
	(*Amat)->ptr         = NULL;
	(*Amat)->row         = NULL;
	(*Amat)->col         = NULL;
	(*Amat)->index       = NULL;
	(*Amat)->bindex		 = NULL;
	(*Amat)->bptr		 = NULL;
	(*Amat)->commtable   = NULL;
	(*Amat)->g_index     = NULL;
	(*Amat)->g_col       = NULL;
	(*Amat)->ranges      = ranges;

	(*Amat)->matrix_type = LIS_MATRIX_NULL;
	(*Amat)->n           = local_n;
	(*Amat)->gn          = global_n;
	(*Amat)->np          = local_n;
	(*Amat)->nnz         = 0;
	(*Amat)->ndz         = 0;
	(*Amat)->nnd         = 0;
	(*Amat)->maxnzr      = 0;
	(*Amat)->bn          = 0;
	(*Amat)->bnr         = 0;
	(*Amat)->bnc         = 0;
	(*Amat)->nr          = 0;
	(*Amat)->nc          = 0;
	(*Amat)->bnnz        = 0;
	(*Amat)->comm        = comm;
	(*Amat)->my_rank     = my_rank;
	(*Amat)->nprocs      = nprocs;
	(*Amat)->is          = is;
	(*Amat)->ie          = ie;

	return LIS_SUCCESS;
}

int lis_matrix_duplicate(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
	int n,np,err;
	int nprocs;
	int i;
	int *ranges;

	if( Ain==NULL )
	{
		SETERR("matrix A is undefined\n");
		return LIS_FAILS;
	}
	if( Ain->matrix_type==LIS_MATRIX_NULL )
	{
		SETERR("matrix A is undefined\n");
		return LIS_FAILS;
	}
	if( Ain->matrix_type<=LIS_MATRIX_ASSEMBLING )
	{
		SETERR("matrix A is assembling\n");
		return LIS_FAILS;
	}

	nprocs = Ain->nprocs;
	*Aout = NULL;

	*Aout = (LIS_MATRIX)lis_malloc( sizeof(struct LIS_MATRIX_STRUCT) );
	if( NULL==*Aout )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	#ifdef _MPI
		ranges = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
		if( ranges==NULL )
		{
			lis_free(*Aout);
			*Aout = NULL;
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		for(i=0;i<nprocs+1;i++) ranges[i] = Ain->ranges[i];
		(*Aout)->ranges      = ranges;
	#else
		(*Aout)->ranges      = NULL;
	#endif

	(*Aout)->value       = NULL;
	(*Aout)->ptr         = NULL;
	(*Aout)->row         = NULL;
	(*Aout)->col         = NULL;
	(*Aout)->index       = NULL;
	(*Aout)->bindex		 = NULL;
	(*Aout)->bptr		 = NULL;
	(*Aout)->commtable   = NULL;
	(*Aout)->g_index     = NULL;
	(*Aout)->g_col       = NULL;

	(*Aout)->matrix_type = LIS_MATRIX_NULL;
	(*Aout)->n           = Ain->n;
	(*Aout)->gn          = Ain->gn;
	(*Aout)->np          = Ain->np;
	(*Aout)->nnz         = 0;
	(*Aout)->ndz         = 0;
	(*Aout)->nnd         = 0;
	(*Aout)->maxnzr      = 0;
	(*Aout)->bn          = 0;
	(*Aout)->bnr         = 0;
	(*Aout)->bnc         = 0;
	(*Aout)->nr          = 0;
	(*Aout)->nc          = 0;
	(*Aout)->bnnz        = 0;
	(*Aout)->comm        = Ain->comm;
	(*Aout)->my_rank     = Ain->my_rank;
	(*Aout)->nprocs      = Ain->nprocs;
	(*Aout)->is          = Ain->is;
	(*Aout)->ie          = Ain->ie;

#ifdef _MPI
	err = lis_commtable_duplicate(Ain,Aout);
	if( err )
	{
		lis_free(ranges);
		lis_free(*Aout);
		*Aout = NULL;
	}
	return err;
#else
	return LIS_SUCCESS;
#endif
}


int lis_matrix_set(int i, int j, LIS_SCALAR value, LIS_MATRIX A)
{
	int n,gn,nnz,np;
	int err;

	n   = A->n;
	gn  = A->gn;
	np  = A->np;
	nnz = A->nnz;
	if( i<0  || j<0 )
	{
		SETERR("i(=%d) or j(=%d) are less than 0\n",i,j);
		return LIS_FAILS;
	}
	if( i>=gn || j>=gn )
	{
		SETERR("i(=%d) or j(=%d) are larger than global n=(%d)\n",i,j,gn);
		return LIS_FAILS;
	}

	if( A->matrix_type==LIS_MATRIX_NULL )
	{
		np = n*5;
		err = LIS_TRUE;
		do
		{
			A->row   = (int *)malloc(np*sizeof(int));
			if( A->row==NULL ) break;
			A->col   = (int *)malloc(np*sizeof(int));
			if( A->col==NULL ) break;
			A->value = (LIS_SCALAR *)malloc(np*sizeof(LIS_SCALAR));
			if( A->value==NULL ) break;
			err = LIS_FALSE;
		}while(LIS_FALSE);
		if( err )
		{
			lis_matrix_destroy(A);
			A = NULL;
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		A->np          = np;
		A->matrix_type = LIS_MATRIX_ASSEMBLING;
	}
	if( nnz==np )
	{
		np      += NNZ_BLOCK;
		err = LIS_TRUE;
		do
		{
			A->row   = (int *)realloc(A->row, np*sizeof(int));
			if( A->row==NULL ) break;
			A->col   = (int *)realloc(A->col, np*sizeof(int));
			if( A->col==NULL ) break;
			A->value = (LIS_SCALAR *)realloc(A->value, np*sizeof(LIS_SCALAR));
			if( A->value==NULL ) break;
			err = LIS_FALSE;
		}while(LIS_FALSE);
		if( err )
		{
			lis_matrix_destroy(A);
			A = NULL;
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		A->np  = np;
	}
	A->row[nnz]   = i;
	A->col[nnz]   = j;
	A->value[nnz] = value;
	A->nnz++;
	return LIS_SUCCESS;
}

int lis_matrix_get_range(LIS_MATRIX A, int *is, int *ie)
{
	if( A==NULL )
	{
		SETERR("matrix A is undefined\n");
		return LIS_FAILS;
	}
	*is = A->is;
	*ie = A->ie;

	return LIS_SUCCESS;
}

int lis_matrix_sets(int i[], int j[], int value[], LIS_MATRIX A)
{
	return LIS_SUCCESS;
}

int lis_matrix_assemble(LIS_MATRIX *A, int matrix_type, ...)
{
	int			err;
	LIS_MATRIX	B;
	va_list		vvlist;
	int			bnr,bnc;
	int			*row,*col;

	row = NULL;
	col = NULL;
	va_start( vvlist, matrix_type );
	bnr = va_arg(vvlist, int);
	bnc = va_arg(vvlist, int);
	row = va_arg(vvlist, int *);
	col = va_arg(vvlist, int *);
	va_end(vvlist);

	if( *A==NULL )
	{
		SETERR("matrix is undefined\n");
		return LIS_FAILS;
	}
	if( (*A)->matrix_type==LIS_MATRIX_NULL )
	{
		SETERR("matrix is undefined\n");
		return LIS_FAILS;
	}
	if( (*A)->matrix_type==LIS_MATRIX_ASSEMBLING )
	{
		(*A)->matrix_type = LIS_MATRIX_COO;
#ifdef _MPI
		err = lis_matrix_g2l(*A);
		if( err ) return err;
#endif
		err = lis_matrix_convert(*A,&B,matrix_type,bnr,bnc,row,col);
		if( err ) return err;
		lis_matrix_destroy(*A);
		*A = B;
		return LIS_SUCCESS;
	}
	else if( (*A)->matrix_type<0 )
	{
		(*A)->matrix_type = -(*A)->matrix_type;
	}
#ifdef _MPI
	if( (*A)->commtable==NULL )
	{
		err = lis_matrix_g2l(*A);
		if( err ) return err;
	}
#endif
	return LIS_SUCCESS;
}

int lis_matrix_set_crs(int nnz, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_CRS;
	A->nnz         = nnz;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;
	return LIS_SUCCESS;
}

int lis_matrix_set_ccs(int nnz, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_CCS;
	A->nnz         = nnz;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;
	return LIS_SUCCESS;
}

int lis_matrix_set_msr(int nnz, int ndz, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_MSR;
	A->nnz         = nnz;
	A->ndz         = ndz;
	A->index       = index;
	A->value       = value;
	return LIS_SUCCESS;
}

#ifdef _OPENMP
int lis_matrix_set_dia(int *nnd, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int i,maxnnd,nprocs;

	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	maxnnd = 0;
	nprocs = omp_get_max_threads();
	for(i=0;i<nprocs;i++)
	{
		if( maxnnd<nnd[i] ) maxnnd = nnd[i];
	}

	A->matrix_type = -LIS_MATRIX_DIA;
	A->nnd         = maxnnd;
	A->col         = nnd;
	A->index       = index;
	A->value       = value;
	return LIS_SUCCESS;
}
#else
int lis_matrix_set_dia(int nnd, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_DIA;
	A->nnd         = nnd;
	A->index       = index;
	A->value       = value;
	return LIS_SUCCESS;
}
#endif

int lis_matrix_set_ell(int maxnzr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_ELL;
	A->maxnzr      = maxnzr;
	A->index       = index;
	A->value       = value;
	return LIS_SUCCESS;
}

#ifdef _OPENMP
int lis_matrix_set_jds(int nnz, int *maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int maxmaxnzr,i,nprocs;

	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	maxmaxnzr = 0;
	nprocs    = omp_get_max_threads();
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
	}

	A->matrix_type = -LIS_MATRIX_JDS;
	A->nnz         = nnz;
	A->maxnzr      = maxmaxnzr;
	A->col         = maxnzr;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;
	return LIS_SUCCESS;
}
#else
int lis_matrix_set_jds(int nnz, int maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_JDS;
	A->nnz         = nnz;
	A->maxnzr      = maxnzr;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;
	return LIS_SUCCESS;
}
#endif

int lis_matrix_set_bsr(int bnr, int bnc, int bnnz, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_BSR;
	A->bnnz        = bnnz;
	A->nr          = (A->n-1)/bnr+1;
	A->nc          = (A->n-1)/bnc+1;
	A->bnr         = bnr;
	A->bnc         = bnc;
	A->bptr        = bptr;
	A->bindex      = bindex;
	A->value       = value;
	return LIS_SUCCESS;
}

int lis_matrix_set_bsc(int bnr, int bnc, int bnnz, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_BSC;
	A->bnnz        = bnnz;
	A->nr          = (A->n-1)/bnr+1;
	A->nc          = (A->n-1)/bnc+1;
	A->bnr         = bnr;
	A->bnc         = bnc;
	A->bptr        = bptr;
	A->bindex      = bindex;
	A->value       = value;
	return LIS_SUCCESS;
}

int lis_matrix_set_vbr(int nnz, int nr, int nc, int bnnz, int *row, int *col, int *ptr, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_VBR;
	A->nnz         = nnz;
	A->nr          = nr;
	A->nc          = nc;
	A->bnnz        = bnnz;
	A->row         = row;
	A->col         = col;
	A->ptr         = ptr;
	A->bptr        = bptr;
	A->bindex      = bindex;
	A->value       = value;
	return LIS_SUCCESS;
}

int lis_matrix_set_coo(int nnz, int *row, int *col, LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_COO;
	A->nnz         = nnz;
	A->row         = row;
	A->col         = col;
	A->value       = value;
	return LIS_SUCCESS;
}

int lis_matrix_set_dns(LIS_SCALAR *value, LIS_MATRIX A)
{
	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type!=LIS_MATRIX_NULL )
	{
		SETERR("matrix is defined\n");
		return LIS_FAILS;
	}

	A->matrix_type = -LIS_MATRIX_DNS;
	A->value       = value;
	return LIS_SUCCESS;
}

int lis_matrix_create2(LIS_MATRIX_PARAMS params, LIS_MATRIX *Amat)
{
	int err;
	size_t len,ilen;
	int gn,np,n,nnz,ndz,nnd,maxnzr,bnnz,bnr,bnc,bn,nr,nc;
	int matrix_type;
	LIS_MATRIX B;

	*Amat = NULL;

	matrix_type = params.matrix_type;
	gn	        = params.gn;
	np	        = params.np;
	n	        = params.n;
	nnz         = params.nnz;
	ndz         = params.ndz;
	nnd         = params.nnd;
	maxnzr      = params.maxnzr;
	bn          = params.bn;
	bnr         = params.bnr;
	bnc         = params.bnc;
	nr          = params.nr;
	nc          = params.nc;
	bnnz        = params.bnnz;

	if( 0>=n ) return LIS_DIMENSION_LE_ZERO;

	switch( matrix_type )
	{
	case LIS_MATRIX_CRS:
	case LIS_MATRIX_CCS:
		if( 0>=nnz ) return LIS_FAILS;
		break;
	case LIS_MATRIX_MSR:
		if( 0>=nnz ) return LIS_FAILS;
		if( 0>ndz  ) return LIS_FAILS;
		break;
	case LIS_MATRIX_COO:
		if( 0>=nnz ) return LIS_FAILS;
		break;
	case LIS_MATRIX_DIA:
		if( 0>=nnd ) return LIS_FAILS;
		break;
	case LIS_MATRIX_JDS:
		if( 0>=nnz || 0>=maxnzr ) return LIS_FAILS;
		break;
	case LIS_MATRIX_ELL:
		if( 0>=maxnzr ) return LIS_FAILS;
		break;
	case LIS_MATRIX_VBR:
		if( 0>=nnz || 0>=bnnz || 0>=nr || 0>=nc ) return LIS_FAILS;
		break;
	case LIS_MATRIX_BSR:
	case LIS_MATRIX_BSC:
		if( 0>=bnr || 0>=bnc || 0>=bnnz ) return LIS_FAILS;
		nr = (n-1)/bnr + 1;
		nc = (n-1)/bnc + 1;
		break;
	case LIS_MATRIX_DNS:
		break;
	default:
		return LIS_FAILS;
		break;
	}

	*Amat = lis_malloc( sizeof(struct LIS_MATRIX_STRUCT) );
	if( NULL==*Amat ) return LIS_OUT_OF_MEMORY;

	(*Amat)->value       = NULL;
	(*Amat)->ptr         = NULL;
	(*Amat)->row         = NULL;
	(*Amat)->col         = NULL;
	(*Amat)->index       = NULL;
	(*Amat)->bindex      = NULL;
	(*Amat)->bptr        = NULL;
	(*Amat)->commtable   = NULL;
	(*Amat)->g_index     = NULL;
	(*Amat)->g_col       = NULL;
	(*Amat)->ranges      = NULL;

	(*Amat)->matrix_type = matrix_type;
	(*Amat)->is          = 0;
	(*Amat)->ie          = n;
	(*Amat)->gn          = gn;
	(*Amat)->np          = np;
	(*Amat)->n           = n;
	(*Amat)->nnz         = nnz;
	(*Amat)->ndz         = ndz;
	(*Amat)->nnd         = nnd;
	(*Amat)->maxnzr      = maxnzr;
	(*Amat)->bn          = bn;
	(*Amat)->bnr         = bnr;
	(*Amat)->bnc         = bnc;
	(*Amat)->nr          = nr;
	(*Amat)->nc          = nc;
	(*Amat)->bnnz        = bnnz;
	(*Amat)->nprocs      = 1;
	(*Amat)->my_rank     = 0;


	err = LIS_TRUE;
	switch( matrix_type )
	{
	case LIS_MATRIX_CRS:
		(*Amat)->ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
		if( NULL==(*Amat)->ptr ) break;
		(*Amat)->index = (int *)lis_malloc( nnz*sizeof(int) );
		if( NULL==(*Amat)->index ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_CCS:
		(*Amat)->ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
		if( NULL==(*Amat)->ptr ) break;
		(*Amat)->index = (int *)lis_malloc( nnz*sizeof(int) );
		if( NULL==(*Amat)->index ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_MSR:
		(*Amat)->index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
		if( NULL==(*Amat)->index ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_DIA:
		(*Amat)->index = (int *)lis_malloc( nnd*sizeof(int) );
		if( NULL==(*Amat)->index ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( np*nnd*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_ELL:
		(*Amat)->index = (int *)lis_malloc( n*maxnzr*sizeof(int) );
		if( NULL==(*Amat)->index ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( n*maxnzr*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_JDS:
		(*Amat)->row = (int *)lis_malloc( n*sizeof(int) );
		if( NULL==(*Amat)->row ) break;
		(*Amat)->ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
		if( NULL==(*Amat)->ptr ) break;
		(*Amat)->index = (int *)lis_malloc( nnz*sizeof(int) );
		if( NULL==(*Amat)->index ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_BSR:
		(*Amat)->bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
		if( NULL==(*Amat)->bptr ) break;
		(*Amat)->bindex = (int *)lis_malloc( bnnz*sizeof(int) );
		if( NULL==(*Amat)->bindex ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( bnnz*bnr*bnc*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_BSC:
		(*Amat)->bptr = (int *)lis_malloc( (nc+1)*sizeof(int) );
		if( NULL==(*Amat)->bptr ) break;
		(*Amat)->bindex = (int *)lis_malloc( bnnz*sizeof(int) );
		if( NULL==(*Amat)->bindex ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( bnnz*bnr*bnc*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_VBR:
		(*Amat)->row = (int *)lis_malloc( (nr+1)*sizeof(int) );
		if( NULL==(*Amat)->row ) break;
		(*Amat)->col = (int *)lis_malloc( (nc+1)*sizeof(int) );
		if( NULL==(*Amat)->col ) break;
		(*Amat)->ptr = (int *)lis_malloc( (bnnz+1)*sizeof(int) );
		if( NULL==(*Amat)->ptr ) break;
		(*Amat)->bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
		if( NULL==(*Amat)->bptr ) break;
		(*Amat)->bindex = (int *)lis_malloc( bnnz*sizeof(int) );
		if( NULL==(*Amat)->bindex ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_COO:
		(*Amat)->row = (int *)lis_malloc( nnz*sizeof(int) );
		if( NULL==(*Amat)->row ) break;
		(*Amat)->col = (int *)lis_malloc( nnz*sizeof(int) );
		if( NULL==(*Amat)->col ) break;
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	case LIS_MATRIX_DNS:
		(*Amat)->value = (LIS_SCALAR *)lis_malloc( n*n*sizeof(LIS_SCALAR) );
		if( NULL==(*Amat)->value ) break;
		err = LIS_FALSE;
		break;
	}

	if( err )
	{
		lis_matrix_destroy(*Amat);
		return LIS_OUT_OF_MEMORY;
	}

	return LIS_SUCCESS;
}

int lis_matrix_destroy(LIS_MATRIX Amat)
{
	if( Amat )
	{
		if( Amat->row ) lis_free( Amat->row );
		if( Amat->col ) lis_free( Amat->col );
		if( Amat->ptr ) lis_free( Amat->ptr );
		if( Amat->index) lis_free( Amat->index );
		if( Amat->bptr ) lis_free( Amat->bptr );
		if( Amat->bindex ) lis_free( Amat->bindex );
		if( Amat->value ) lis_free( Amat->value );
		if( Amat->ranges ) lis_free( Amat->ranges );
		if( Amat->g_index ) lis_free( Amat->g_index );
		if( Amat->g_col ) lis_free( Amat->g_col );
		if( Amat->commtable ) lis_commtable_destroy( Amat->commtable );
		lis_free(Amat);
	}

	return LIS_SUCCESS;
}

